package JointFeldmanProtocol;

import com.google.protobuf.InvalidProtocolBufferException;
import com.google.protobuf.Message;
import meerkat.protobuf.DKGMessages.*;
import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;
/**
 * Created by Tzlil on 2/7/2016.
 * JointFeldamn protocol assumes all parties can communicate throw broadcast chanel
 * and private chanel (for each pair)
 * this class simulates it
 */
public class Network {

    private final User[] users;
    private final int n;
    private final Queue<Integer> availableIDs;


    public Network(int n) {
        this.n = n;
        this.users = new User[n];
        this.availableIDs = new ArrayBlockingQueue<Integer>(n);
        for (int id = 1; id <= n; id++){
            availableIDs.add(id);
        }
    }

    public User connect(MailHandler messageHandler){
        Integer id = availableIDs.poll();
        if (id == null)
            return null;
        users[id - 1] = new User(id,messageHandler);
        return users[id - 1];
    }

    private boolean sendMessage(User sender,int destination,Mail.Type type,Message message){
        if(destination < 1 || destination > n)
            return false;
        User user = users[destination - 1];
        if (user == null)
            return false;
        Mail mail = Mail.newBuilder()
                .setSender(sender.getID())
                .setDestination(destination)
                .setIsPrivate(true)
                .setType(type)
                .setMessage(message.toByteString())
                .build();
        return user.mailbox.add(mail);
    }

    private void sendBroadcast(User sender,Mail.Type type,Message message){
        User user;
        int ID = sender.ID;
        Mail mail = Mail.newBuilder()
                .setSender(sender.getID())
                .setDestination(0)
                .setIsPrivate(false)
                .setType(type)
                .setMessage(message.toByteString())
                .build();
        for (int i = 0 ; i < n ; i++){
            if (i + 1 == ID) {
                continue;
            }
            user = users[i];
            user.mailbox.add(mail);
        }
    }

    public class User{
        private final MailHandler messageHandler;
        private final Queue<Mail> mailbox;
        private final int ID;
        private final Thread receiverThread;
        public User(int ID, MailHandler messageHandler) {
            this.mailbox = new ArrayBlockingQueue<Mail>(1 << n);
            this.ID = ID;
            this.messageHandler = messageHandler;
            this.receiverThread = new Thread(new Receiver());
        }

        public boolean send(int id, Mail.Type type,Message message){
            return sendMessage(this,id,type,message);
        }
        public void broadcast(Mail.Type type,Message message){
            sendBroadcast(this,type,message);
        }
        public int getID() {
            return ID;
        }
        public Thread getReceiverThread(){
            return receiverThread;
        }
        private class Receiver implements Runnable{
            @Override
            public void run() {
                while (true){
                    if (!mailbox.isEmpty()){
                        try {
                            messageHandler.handel(mailbox.poll());
                        } catch (InvalidProtocolBufferException e) {
                            e.printStackTrace();
                        }
                    }else{
                        try {
                            Thread.sleep(30);
                        } catch (InterruptedException e) {
                            // do nothing
                        }
                    }
                }
            }
        }


    }

    public interface MailHandler {
        public void handel(Mail mail) throws InvalidProtocolBufferException;
    }
}
